/*
    ZSpace - kernels.
    (c) Claus Wimmer 2021
*/

#include "types.h"

__kernel void init(__global t_pixel *pixels) 
{
    __global t_pixel *p = pixels + I;

    clear_vector(&p->e);
    clear_vector(&p->p);
    clear_vector(&p->me);
    clear_vector(&p->mp);
    clear_vector(&p->te);
    clear_vector(&p->tp);
    clear_vector(&p->dex);
    clear_vector(&p->dey);
    clear_vector(&p->dpx);
    clear_vector(&p->dpy);
    p->a = 1.0f;
    p->b = 0.0f;
    p->debug = 0.0f;
}

__kernel void proton_same(__global t_pixel *pixels, __global t_vector *pattern)
{
    __global t_pixel *p = pixels + I;

    p->p.a = pattern->a;
    p->p.b = pattern->b;
    p->p.c = pattern->c;
}

__kernel void electron(__global t_pixel *pixels, __global t_particle *particles, __global int *index)
{
    __global t_pixel *p = pixels + I;

    int x = particles[*index].x + WIDTH / 2, y = particles[*index].y + HEIGHT / 2;
    float 
        RADIUS = 10.0f,
        radius = sqrt(pow((float)X - x, 2.0f) + pow((float)Y - y, 2.0f)),
        phi = atan2((float)Y - y, (float)X - x) + particles[*index].phi,
        corpus = PI * radius / RADIUS/ 2.0f;

    if(PI / 2.0f > corpus)
    {
        p->e.a += cos(phi) * sin(corpus) * particles[*index].polarity;
        p->e.b += sin(phi) * sin(corpus);
        p->e.c += cos(corpus) * particles[*index].spin;
    }
    else
    {
        p->e.a += PI / 2.0f / corpus * cos(phi) * particles[*index].polarity;
        p->e.b += PI / 2.0f / corpus * sin(phi);
    }
}

__kernel void proton(__global t_pixel *pixels, __global t_particle *particles, __global int *index)
{
    __global t_pixel *p = pixels + I;

    int x = particles[*index].x + WIDTH / 2, y = particles[*index].y + HEIGHT / 2;
    float 
        RADIUS = 10.0f,
        radius = sqrt(pow((float)X - x, 2.0f) + pow((float)Y - y, 2.0f)),
        phi = atan2((float)Y - y, (float)X - x) + particles[*index].phi,
        corpus = PI * radius / RADIUS/ 2.0f;

    if(PI / 2.0f > corpus)
    {
        p->p.a += cos(phi) * sin(corpus) * particles[*index].polarity;
        p->p.b += sin(phi) * sin(corpus);        
        p->p.c += cos(corpus) * particles[*index].spin;
    }
    else
    {
        p->p.a += PI / 2.0f / corpus * cos(phi) * particles[*index].polarity;
        p->p.b += PI / 2.0f / corpus * sin(phi);
    }
}

__kernel void norm(__global t_pixel *pixels) 
{
    __global t_pixel *p = pixels + I;

    norm_vector(&p->e);
    norm_vector(&p->p);
}

__kernel void derive1(__global t_pixel * pixels)
{
    __global t_pixel *p = pixels + I, *l = p - 1, *u = p - WIDTH;

    if(0 < X && 0 < Y)
    {
        // e
        crossProd_vector(&p->dex, &l->e, &p->e);
        crossProd_vector(&p->dey, &u->e, &p->e);
        // p
        crossProd_vector(&p->dpx, &l->p, &p->p);
        crossProd_vector(&p->dpy, &u->p, &p->p);
     }
}

__kernel void derive2(__global t_pixel *pixels)
{
    __global t_pixel *p = pixels + I, *r = p + 1, *d = p + WIDTH;

    clear_vector(&p->te);
    clear_vector(&p->tp);
    
    if(WIDTH - 1 > X && HEIGHT - 1 > Y)
    {        
        // e
        multAdd_vector(+L_F, &p->te, &p->dex);
        multAdd_vector(-L_F, &p->te, &r->dex);
        multAdd_vector(+L_F, &p->te, &p->dey);
        multAdd_vector(-L_F, &p->te, &d->dey);
        // p
        multAdd_vector(+L_F, &p->tp, &p->dpx);
        multAdd_vector(-L_F, &p->tp, &r->dpx);
        multAdd_vector(+L_F, &p->tp, &p->dpy);
        multAdd_vector(-L_F, &p->tp, &d->dpy);
    }
    // Radius.
    multAdd_vector(1.0f - get_length_vector(&p->e), &p->me, &p->e);
    multAdd_vector(1.0f - get_length_vector(&p->p), &p->mp, &p->p);
    // Corpus.
    p->me.c -= RESTORING_MOMENT_E * p->e.c;
    p->mp.c -= RESTORING_MOMENT_P * p->p.c;
    // Turn laplace.
    addTurn_vector(&p->me, &p->e, &p->te);
    addTurn_vector(&p->mp, &p->p, &p->tp);
}

__kernel void crosstalk(__global t_pixel *pixels)
{
    __global t_pixel *p = pixels + I, *r = p + 1, *d = p + WIDTH;

    if(0 < X && 0 < Y && WIDTH - 1 > X && HEIGHT - 1 > Y)
    {
        // Corpus
        float
            cex = p->dex.c + (p + 1)->dex.c,
            cexu = (p - WIDTH)->dex.c + (p - WIDTH + 1)->dex.c + cex,
            cexd = (p + WIDTH)->dex.c + (p + WIDTH + 1)->dex.c + cex,
            cey = p->dey.c + (p + WIDTH)->dey.c,
            ceyl = (p - 1)->dey.c + (p - 1 + WIDTH)->dey.c + cey,
            ceyr = (p + 1)->dey.c + (p + WIDTH + 1)->dey.c + cey,
            cpx = p->dpx.c + (p + 1)->dpx.c,
            cpxu = (p - WIDTH)->dpx.c + (p - WIDTH + 1)->dpx.c + cpx,
            cpxd = (p + WIDTH)->dpx.c + (p + WIDTH + 1)->dpx.c + cpx,
            cpy = p->dpy.c + (p + WIDTH)->dpy.c,
            cpyl = (p - 1)->dpy.c + (p - 1 + WIDTH)->dpy.c + cpy,
            cpyr = (p + 1)->dpy.c + (p + WIDTH + 1)->dpy.c + cpy;
        // Drehen erfolgt hier falsch, muss add-turn heißen.
        // p => e
        /* pdx */ p->e.c -= O_F * (p->dpx.c * ceyl - r->dpx.c * ceyr);
        /* pdy */ p->e.c -= O_F * (p->dpy.c * cexu - d->dpy.c * cexd);
        // e => p
        /* edx */ p->p.c -= O_F * (p->dex.c * cpyl - r->dex.c * cpyr);
        /* edy */ p->p.c -= O_F * (p->dey.c * cpxu - d->dey.c * cpxd);
    } 
}

__kernel void turn(__global t_pixel *pixels)
{
    __global t_pixel *p = pixels + I;

    if((0 < X) && (0 < Y) && (WIDTH - 1 > X) && (HEIGHT - 1 > Y))
    {
        // Electron.
        //saturate_vector(&p->e);
        multAdd_vector(1.0f, &p->e, &p->me);
        //mult_vector(0.999f, &p->me);
        // Proton.
        //saturate_vector(&p->p);
        multAdd_vector(1.0f, &p->p, &p->mp);
        //mult_vector(0.999f, &p->mp);        
    } // End if - no border -.
    // Border region.
    if(0==X)
    {
        p->e = (p + 1)->e;
        p->p = (p + 1)->p;
    }
    if(0 == Y)
    {
        p->e = (p + WIDTH)->e;
        p->p = (p + WIDTH)->p;
    }
    if(WIDTH - 1 == X)
    {
        p->e = (p - 1)->e;
        p->p = (p - 1)->p;
    }
    if(HEIGHT - 1 == Y)
    {
        p->e = (p - WIDTH)->e;
        p->p = (p - WIDTH)->p;
    }
}

__kernel void render(__global t_pixel *pixels, __global t_rgb *rgbs, __global int *zoom_in)
{
    int xp = X - X % ARROW_SIZE + ARROW_SIZE / 2, yp = Y - Y % ARROW_SIZE + ARROW_SIZE / 2;
    __global t_pixel 
        *p = pixels + (Y / *zoom_in) * (WIDTH / *zoom_in) + (X / *zoom_in),
        *p1 = pixels + (yp / *zoom_in) * (WIDTH / *zoom_in) + (xp / *zoom_in);
    __global t_rgb *rgb = rgbs + Y * WIDTH + X;
    
    //rgb->red = rgb->green = rgb->blue = saturate_rgb(127.0f * (1.0f + 1e0f * p->e.c));

    rgb->red = saturate_rgb(255.0f * 1e+0f * fabs(p->e.c));
    rgb->green = saturate_rgb(255.0f * 1e+0f * fabs(p->p.c));
    rgb->blue = 0 * saturate_rgb(255.0f * 1e4f * fabs(p->debug));

    {
        float 
            red_len = sqrt(pow(p1->e.a, 2.0f) + pow(p1->e.b, 2.0f)),
            red1 = line(xp, yp, X, Y, p1->e.a, p1->e.b, E_LEN * red_len, T_LINE_E),                 
            red2 = line(xp - E_LEN * p1->e.a, yp - E_LEN * p1->e.b, X, Y, p1->e.b, -p1->e.a, T_MARKER * red_len, T_LINE_E),
            green_len = sqrt(pow(p1->p.a, 2.0f) + pow(p1->p.b, 2.0f)),
            green_1 = line(xp + T_MARKER * p1->p.b - (T_MARKER * sqrt(0.5f) * 0.25f + 0.5f * T_LINE_P) * p1->p.a, yp - T_MARKER * p1->p.a - (T_MARKER * sqrt(0.5f) * 0.25f + 0.5f * T_LINE_P) * p1->p.b, X, Y, p1->p.a, p1->p.b, (P_LEN - T_MARKER * sqrt(0.5f) * 0.25f) * green_len, T_LINE_P),
            green_2 = line(xp - T_MARKER * p1->p.b - (T_MARKER * sqrt(0.5f) * 0.25f + 0.5f * T_LINE_P) * p1->p.a, yp + T_MARKER * p1->p.a - (T_MARKER * sqrt(0.5f) * 0.25f + 0.5f * T_LINE_P) * p1->p.b, X, Y, p1->p.a, p1->p.b, (P_LEN - T_MARKER * sqrt(0.5f) * 0.25f) * green_len, T_LINE_P),
            green_3 = line(xp + P_LEN * p1->p.a + 0.5f * T_MARKER * p1->p.b, yp + P_LEN * p1->p.b - 0.5f * T_MARKER * p1->p.a, X, Y, -p1->p.b + p1->p.a, p1->p.a + p1->p.b, T_MARKER * green_len * sqrt(0.5f), T_LINE_P),
            green_4 = line(xp + P_LEN * p1->p.a - 0.5f * T_MARKER * p1->p.b, yp + P_LEN * p1->p.b + 0.5f * T_MARKER * p1->p.a, X, Y, -p1->p.b - p1->p.a, p1->p.a - p1->p.b, T_MARKER * green_len * sqrt(0.5f), T_LINE_P),
            green_5 = line(xp - P_LEN * p1->p.a, yp - P_LEN * p1->p.b, X, Y, p1->p.b, -p1->p.a, T_MARKER * green_len, T_LINE_P),
            green = green_1 + green_2 + green_3 + green_4 + green_5, 
            blue_len = sqrt(pow(p1->a, 2.0f) + pow(p1->b, 2.0f)),
            blue = line(xp, yp, X, Y, p1->a, p1->b, E_LEN * blue_len, T_LINE_E),
            red = red1 + red2,
            alpha = red;

        green = 1.0f < green ? 1.0f : green; 
        red = 1.0f < red ? 1.0f : red; 

        if(green > alpha) alpha = green;
        if(blue > alpha) alpha = blue;

        rgb->red *= 1.0f - alpha;
        rgb->green *= 1.0f - alpha;
        rgb->blue *= 1.0f - alpha;

        rgb->red += red * 255.0f; 
        rgb->green += green * 255.0f;
        rgb->blue += blue * 255.0f; 
    }
}
